import os
replace_char = "\r\n" if os.name in ("nt", ) else "\n"

# ---------------------------- README.rst ---------------------------------- #
default_read_me_for_user = """
madliar

  This is a tiny WSGI web application freamwork. For more infomation refers to 
https://github.com/cl-ei/madliar/blob/master/README.rst

  Thanks for checking it out.

""".replace("\n", replace_char)


# ---------------------- application/urls.py ------------------------------- #
default_user_url_map = """
from madliar.http.response import HttpResponse


# The `hello_world` function is just used to show you how to start an application,
# you may need remove it and build your own url map latter.
def hello_world(request):
    return HttpResponse("Hello world !")


url = {
    "/": hello_world,

}
""".replace("\n", replace_char)


# --------------- management/madliar_uwsgi_socket.xml ---------------------- #
default_wsgi_xml = """
<uwsgi>
    <socket>:65501</socket>
    <chdir>%s</chdir>
    <module>management.wsgi</module>
    <processes>1</processes>
    <daemonize>%s_madliar.log</daemonize>
</uwsgi>
""".replace("\n", replace_char)


# --------------------- management/wsgi.py --------------------------------- #
default_wsgi_py = """
\"""
The public interface to WSGI support. It returns a WSGI callable.

\"""

from madliar.core import get_application


application = get_application()
""".replace("\n", replace_char)


# --------------------- management/config.py --------------------------------- #
default_user_config = """\"""
madliar freamwork settings.

\"""

# If set True, An internal static files distribution server will be started,
# and more detail info will be provided when error occured.
DEBUG = True

# Enable the debug log output.
ENABLE_SYS_LOG = True
SYS_LOG_PATH = "./"

# Statics files response url. It will ignored when DEBUG=False
STATICS_URL_MAP = {

}

# To load middleware.
INSTALLED_MIDDLEWARE = (

)
""".replace("\n", replace_char)



import os
from madliar.config import settings


def make_logger(name, log_file_path, level="DEBUG", log_format=None):
    import logging
    if not log_format:
        log_format = "%(levelname)s %(asctime)s: %(message)s"

    level_names = {
        "CRITICAL": logging.CRITICAL,
        "ERROR": logging.ERROR,
        "WARN": logging.WARNING,
        "WARNING": logging.WARNING,
        "INFO": logging.INFO,
        "DEBUG": logging.DEBUG,
        "NOTSET": logging.NOTSET,
    }

    fh = logging.FileHandler(log_file_path)
    fh.setFormatter(logging.Formatter(log_format))

    logger = logging.getLogger(name)
    logger.setLevel(level_names.get(level.upper(), logging.DEBUG))
    logger.addHandler(fh)
    return logger


logging = make_logger(
    name="madliar",
    log_file_path=os.path.join(settings.SYS_LOG_PATH, "madliar_sys.log"),
    level="INFO",
)



import re
import os
from wsgiref.headers import Headers
from madliar import __version__


STATICS_FILE_MIME_TYPE = (
    ("xml",             "text/xml"),
    ("css",             "text/css"),
    ("html htm shtml",  "text/html"),
    ("txt",             "text/plain"),
    ("mml",             "text/mathml"),
    ("wml",             "text/vnd.wap.wml"),
    ("htc",             "text/x-component"),
    ("jad",             "text/vnd.sun.j2me.app-descriptor"),

    ("png",             "image/png"),
    ("gif",             "image/gif"),
    ("jpeg jpg",        "image/jpeg"),
    ("tif tiff",        "image/tiff"),
    ("webp",            "image/webp"),
    ("jng",             "image/x-jng"),
    ("ico",             "image/x-icon"),
    ("svg svgz",        "image/svg+xml"),
    ("bmp",             "image/x-ms-bmp"),
    ("wbmp",            "image/vnd.wap.wbmp"),

    ("mp4",             "video/mp4"),
    ("ts",              "video/mp2t"),
    ("mpeg mpg",        "video/mpeg"),
    ("3gpp 3gp",        "video/3gpp"),
    ("webm",            "video/webm"),
    ("mng",             "video/x-mng"),
    ("m4v",             "video/x-m4v"),
    ("flv",             "video/x-flv"),
    ("wmv",             "video/x-ms-wmv"),
    ("asx asf",         "video/x-ms-asf"),
    ("avi",             "video/x-msvideo"),
    ("mov",             "video/quicktime"),

    ("ogg",             "audio/ogg"),
    ("mp3",             "audio/mpeg"),
    ("mid midi kar",    "audio/midi"),
    ("m4a",             "audio/x-m4a"),
    ("ra",              "audio/x-realaudio"),

    ("js",              "application/javascript"),
    ("run",             "application/x-makeself"),
    ("xls",             "application/vnd.ms-excel"),
    ("jardiff",         "application/x-java-archive-diff"),
    ("rar",             "application/x-rar-compressed"),
    ("xpi",             "application/x-xpinstall"),
    ("sea",             "application/x-sea"),
    ("hqx",             "application/mac-binhex40"),
    ("sit",             "application/x-stuffit"),
    ("rtf",             "application/rtf"),
    ("kml",             "application/vnd.google-earth.kml+xml"),
    ("xhtml",           "application/xhtml+xml"),
    ("jnlp",            "application/x-java-jnlp-file"),
    ("ppt",             "application/vnd.ms-powerpoint"),
    ("atom",            "application/atom+xml"),
    ("m3u8",            "application/vnd.apple.mpegurl"),
    ("rss",             "application/rss+xml"),
    ("cco",             "application/x-cocoa"),
    ("jar war ear",     "application/java-archive"),
    ("tcl tk",          "application/x-tcl"),
    ("prc pdb",         "application/x-pilot"),
    ("woff",            "application/font-woff"),
    ("zip",             "application/zip"),
    ("doc",             "application/msword"),
    ("eot",             "application/vnd.ms-fontobject"),
    ("kmz",             "application/vnd.google-earth.kmz"),
    ("ps eps ai",       "application/postscript"),
    ("json",            "application/json"),
    ("pdf",             "application/pdf"),
    ("pl pm",           "application/x-perl"),
    ("7z",              "application/x-7z-compressed"),
    ("der pem crt",     "application/x-x509-ca-cert"),
    ("xspf",            "application/xspf+xml"),
    ("swf",             "application/x-shockwave-flash"),
    ("wmlc",            "application/vnd.wap.wmlc"),
    ("rpm",             "application/x-redhat-package-manager"),
    ("xlsx", "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet"),
    ("docx", "application/vnd.openxmlformats-officedocument.wordprocessingml.document"),
    ("pptx", "application/vnd.openxmlformats-officedocument.presentationml.presentation"),
    ("bin exe dll deb dmg iso img msi msp msm", "application/octet-stream"),
)


class BaseResponse(object):
    def __init__(self, status_code=None, reason_phrase=None, content_type=None, charset=None):
        try:
            if not status_code:
                status_code = 200
                reason_phrase = "OK"

            self.status_code = int(status_code)
            self.reason_phrase = reason_phrase

        except (TypeError, ValueError):
            # TODO: add traceback
            self.status_code = 500
            self.reason_phrase = "Internal Server Error"

        self.cookies = []
        self.charset = charset or "utf-8"
        self._handler_class = None
        self._content = []
        self.__set_default_headers()

        content_type = "text/html" if not content_type else content_type.strip(" \r\n")
        if (
            content_type.split("/", 1)[0] in ("text", "application")
            and "charset" not in content_type.lower()
        ):
            content_type = "%s; charset=%s" % (content_type.rstrip("; \r\n"), self.charset)
        self.headers.add_header("Content-Type", content_type)

    def __set_default_headers(self):
        self.headers = Headers([
            ("Server", "MadLiar/%s" % __version__),
            ("Access-Control-Allow-Origin", "*"),
            ("X-Frame-Options", "SAMEORIGIN"),
        ])

    @property
    def content(self):
        return b''.join(self._content)

    @content.setter
    def content(self, value):
        # Consume iterators upon assignment to allow repeated iteration.
        if not hasattr(value, '__iter__'):
            value = [value]

        content = b''.join(map(
            lambda x: bytes(x) if isinstance(x, bytes) else bytes(x.encode(self.charset)),
            value
        ))
        self.headers.add_header("Content-length", str(len(content)))
        self._content = [content]

    def __iter__(self):
        return iter(self._content)


class HttpResponse(BaseResponse):
    def __init__(self, content, *args, **kwargs):
        super(HttpResponse, self).__init__(*args, **kwargs)
        if type(content) == unicode:
            content = content.encode(self.charset)
        if type(content) in (bytes, str):
            self.content = content


class Http404Response(BaseResponse):
    def __init__(self, *args, **kwargs):
        BaseResponse.__init__(self, *args, **kwargs)
        self.content = "<center><h3>404 Not Found!</h3></center>"
        self.status_code = 404
        self.reason_phrase = "Not Found"


class Http410Response(BaseResponse):
    def __init__(self, *args, **kwargs):
        BaseResponse.__init__(self, *args, **kwargs)
        self.content = "<center><h3>410 Gone!</h3></center>"
        self.status_code = 410
        self.reason_phrase = "Gone"


class Http500Response(BaseResponse):
    def __init__(self, *args, **kwargs):
        BaseResponse.__init__(self, *args, **kwargs)
        self.content = "<center><h3>500 Internal Server Error!</h3></center>"
        self.status_code = 500
        self.reason_phrase = "Internal Server Error"


class StreamingHttpResponse(BaseResponse):
    def _set_streaming_content(self, value):
        return self, value


class FileResponse(StreamingHttpResponse):
    """
    A streaming HTTP response class optimized for files.
    """
    block_size = 4096

    def _set_streaming_content(self, value):
        super(FileResponse, self)._set_streaming_content(value)


class HttpResponseRedirectBase(HttpResponse):
    allowed_schemes = ['http', 'https', 'ftp']

    def __init__(self, redirect_to, *args, **kwargs):
        super(HttpResponse, self).__init__()
        self._content_type_for_repr = ""
        self._redirect_to = redirect_to

    url = property(lambda self: self['Location'])

    def __repr__(self):
        return '<%(cls)s status_code=%(status_code)d%(content_type)s, url="%(url)s">' % {
            'cls': self.__class__.__name__,
            'status_code': self.status_code,
            'content_type': self._content_type_for_repr,
            'url': self.url,
        }


def static_files_response(request, static_file_path):
    try:
        with open(static_file_path, "rb") as f:
            content = f.read()
    except IOError:
        return Http404Response()

    content_type = "application/octet-stream"
    static_file_ext_name = os.path.splitext(static_file_path)[-1].lower().lstrip(".")
    if static_file_path:
        for file_type, description in STATICS_FILE_MIME_TYPE:
            if static_file_ext_name in file_type:
                content_type = description
                break

    return HttpResponse(content, content_type=content_type)



import os
import sys

from random import randint
from importlib import import_module
from madliar.utils import get_traceback


__all__ = (
    "reg_command",
    "execute_from_command_line",
)

main_help_text = """
 -------------------------------------------------------------------------
    madLiar.

    Usage:
        madliar-manage <command> [options]

    Commands:

        create_proj [project_name]
            Create a project on local path. Specify the name of working directory.

        runserver [host:port]
            Running a simple server for debug. Specify the host and port of server.

        help
            Show help for commands.

    Custom commands:
%s
 -------------------------------------------------------------------------
"""

__madliar_command_attr_key = "__madliar_command_%0x___" % randint(0, 0xFFFFFFFF)


def reg_command(name):
    def wrap(f):
        setattr(f, __madliar_command_attr_key, name)
        return f
    return wrap


def _get_command_name(f):
    return getattr(f, __madliar_command_attr_key, None)


class built_in_command(object):
    CommandDoesNotExisted = type("built_in_command__CommandDoesNotExisted", (Exception, ), {})
    __function = {}

    def __init__(self, command):
        self.__command = command

    def __call__(self, func):
        self.__class__.__function[self.__command] = func
        return func

    @classmethod
    def run(cls, command, *args, **kwargs):
        picked_func = cls.__function.get(command)
        if callable(picked_func):
            return picked_func(*args, **kwargs)
        else:
            raise cls.CommandDoesNotExisted


@built_in_command(command="runserver")
def runserver(*args):
    try:
        host, port = args[0].split(":")
        port = int(port)
    except (IndexError, TypeError, ValueError) as e:
        sys.stdout.write(
            "An error happend: %s, use default configured.\n" % str(e)
        )
        host, port = "0.0.0.0", 8080

    sys.stdout.write("Run madliar wsgi server on %s:%s.\n" % (host, port))
    from madliar.core import wsgi_server
    wsgi_server(host, int(port)).serve_forever()


@built_in_command(command="create_proj")
def create_proj(*args):
    try:
        project_name = args[0]
    except IndexError:
        sys.stderr.write("You must type a folder name.")
        return 0

    import string
    avaliable_name_c = string.letters + string.digits + "_"
    if set(project_name) - set(avaliable_name_c):
        sys.stderr.write("Include special characters.")
        return 0

    try:
        from madliar.config.project_template import (
            default_user_url_map,
            default_wsgi_xml,
            default_wsgi_py,
            default_read_me_for_user,
            default_user_config
        )
        project_folder = os.path.join(os.getcwd(), project_name)
        applicaton_folder = os.path.join(project_folder, "application")
        management_folder = os.path.join(project_folder, "management")

        os.mkdir(project_folder)
        os.mkdir(applicaton_folder)
        os.mkdir(management_folder)

        with open(os.path.join(project_folder, "README.rst"), "ab") as f:
            f.write(default_read_me_for_user)
        with open(os.path.join(applicaton_folder, "__init__.py"), "ab"):
            pass
        with open(os.path.join(applicaton_folder, "urls.py"), "ab") as f:
            f.write(default_user_url_map)
        with open(os.path.join(management_folder, "__init__.py"), "ab"):
            pass
        with open(os.path.join(management_folder, "madliar_uwsgi_socket.xml"), "ab") as f:
            f.write(default_wsgi_xml % (os.path.abspath("."), project_name))
        with open(os.path.join(management_folder, "wsgi.py"), "ab") as f:
            f.write(default_wsgi_py)
        with open(os.path.join(management_folder, "config.py"), "ab") as f:
            f.write(default_user_config)

        sys.stdout.write("Application %s created! " % project_name)
    except Exception as e:
        sys.stderr.write("Application %s not create: %s" % (project_name, e))
        return 0


class ManagementUtility(object):
    """
    Encapsulates the logic of the command tools utilities.

    A ManagementUtility finds out the command in the lib.command.py file
    and run it.
    """
    def __init__(self, argv=None):
        self.argv = argv or sys.argv[:]
        self.prog_name = os.path.basename(self.argv[0])
        self.settings_exception = None

    def show_help_text(self):
        custom_command = ""
        for command, target_fucntion in self.load_custom_command().iteritems():
            custom_command += "\n" + " "*8 + command
            if target_fucntion.__doc__:
                doc_lines = target_fucntion.__doc__.splitlines()
                if doc_lines:
                    command_doc = "".join(
                        ["\n" + " "*12 + line.lstrip(" ") for line in doc_lines]
                    ).rstrip(" \n")
                else:
                    command_doc = "\n            No description."
                custom_command += command_doc + "\n"

        if not custom_command:
            custom_command = (
                "\n        Not found custom command.\n"
                "\n        You can write functions that receive system"
                "\n        arguments and process your own buessess logic,"
                "\n        and put it to `command` directory under your"
                "\n        `application` directory.\n"
            )

        full_help_text = main_help_text % custom_command
        sys.stdout.write(full_help_text)

        return 0

    @staticmethod
    def load_custom_command():
        from madliar.utils import find_all_package

        sys.path.append(".")
        application_path = "application"
        if not os.path.exists(application_path):
            return {}

        all_package = find_all_package(".")
        commands = {}

        for package in all_package:
            try:
                m = import_module(package)
            except Exception as e:
                sys.stderr.write(
                    "\nAn error happend when load custom command: %s.\n%s"
                    "\n\nNow ignore it, remanber to fix it.\n"
                    % (e, get_traceback())
                )
                continue

            public_attrs = [attr for attr in dir(m) if not attr.startswith("_")]
            for attr in public_attrs:
                command_func = getattr(m, attr)
                if not callable(command_func):
                    continue

                command_name = _get_command_name(command_func)
                if not command_name:
                    continue

                if command_name in commands:
                    sys.stderr.write(
                        "Found duplicate command name: \"%s\" in `%s`, "
                        "a unique name is required, please check it."
                        % (command_name, package)
                    )
                    sys.exit(0)

                commands[command_name] = command_func

        return commands

    def _execute_custom_command(self, *args):
        target, t_args,  = args[0], args[1:]

        custom_command = self.load_custom_command()
        target_function = custom_command.get(target)
        if not callable(target_function):
            sys.stderr.write("Unknow command: %s, try typing `help` ?" % target)
            return 0

        try:
            return target_function(*t_args)
        except Exception as e:
            sys.stderr.write(
                "An error happend when excute command `%s`:\n    %s"
                % (target, e)
            )
        return 0

    def execute(self):
        """
        Given the command-line arguments, this figures out which subcommand is
        being run, and runs it.
        """
        if len(self.argv) < 2 or self.argv[1] == "help":
            return self.show_help_text()

        sys.stdout.write("\nmadliar.\n\n")
        try:
            excute_resule = built_in_command.run(*self.argv[1:])
        except built_in_command.CommandDoesNotExisted:
            excute_resule = self._execute_custom_command(*self.argv[1:])

        return excute_resule


def execute_from_command_line(argv=None):
    """
    A simple method that runs a Manager.
    """
    utility = ManagementUtility(argv)
    utility.execute()



"""
Settings and configuration for madliar.

"""

import os
import sys
from madliar import exceptions as mad_exceptions


class SettingsBuilder(type):
    def __new__(mcs, *args):
        name, bases, namespace = args[:]
        try:
            sys.path.append(os.getcwd())
            management = __import__("management.config")
            user_settings = getattr(management, "config")

            slots_dict = {
                slot: getattr(user_settings, slot)
                for slot in dir(user_settings) if not slot.startswith("_")
            }

        except AttributeError:
            sys.stderr.write("Cannot read user's custom settings.")
            raise mad_exceptions.ProjectFolderStructureError(
                "This error occured because the project folder is not " 
                "in accordance with the rules of madliar, see the detail "
                "at `README.rst`."
            )
        except ImportError:
            slots_dict = {}

        # overwrite project default settings
        if "DEBUG" not in slots_dict:
            slots_dict["DEBUG"] = True

        slots_dict["PROJECT_CWD"] = os.path.abspath(".")

        installed_middle_ware = ["madliar.http.middleware.BaseMiddleware"]
        custom_middleware = slots_dict.get("INSTALLED_MIDDLEWARE", [])
        if not isinstance(custom_middleware, (list, tuple)):
            sys.stderr.write(
                "You must regist your middleware in INSTALLED_MIDDLEWARE tuple. not: \n%s"
                % custom_middleware
            )
            sys.exit(0)

        installed_middle_ware.extend(custom_middleware)
        slots_dict["INSTALLED_MIDDLEWARE"] = installed_middle_ware

        if "ENABLE_SYS_LOG" not in slots_dict:
            slots_dict["ENABLE_SYS_LOG"] = True

        if "SYS_LOG_PATH" not in slots_dict:
            slots_dict["SYS_LOG_PATH"] = slots_dict["PROJECT_CWD"]

        if slots_dict["DEBUG"] and "STATICS_URL_MAP" not in slots_dict:
            slots_dict["STATICS_URL_MAP"] = {}

        namespace["__slots__"] = slots_dict.keys()

        def __init__(self):
            for k, v in slots_dict.iteritems():
                setattr(self, k, v)

            def __setattr__(s, key, value):
                raise mad_exceptions.DefaultSettingsWriteError(
                    "Cannot set {%s: %s}, madliar default settings cannot be overwrite."
                    % (key, value)
                )
            self.__class__.__setattr__ = __setattr__

        namespace["__init__"] = __init__
        new_cls = type.__new__(mcs, name, bases, namespace)

        return new_cls


class SettingsLoader(object):
    __metaclass__ = SettingsBuilder


settings = SettingsLoader()



"""
Tools for working with functions and callable objects.

"""
import os

import StringIO
import traceback


class cached_property(object):
    """
    Decorator that converts a method with a single self argument into a
    property cached on the instance.

    Optional ``name`` argument allows you to make cached properties of other
    methods. (e.g.  url = cached_property(get_absolute_url, name='url') )
    """
    def __init__(self, func, name=None):
        self.func = func
        self.__doc__ = getattr(func, '__doc__')
        self.name = name or func.__name__

    def __get__(self, instance, cls=None):
        if instance is None:
            return self
        res = instance.__dict__[self.name] = self.func(instance)
        return res


def get_traceback():
    except_trace = StringIO.StringIO()
    traceback.print_exc(file=except_trace)
    except_content = str(except_trace.getvalue())
    except_trace.close()
    return except_content


def find_all_package(path):
    from setuptools import findall

    path_split_char = "\\" if os.name in ("nt",) else "/"
    all_packages_file = []

    for f in findall(path):
        f_path, f_ex = os.path.splitext(f)
        if f_ex in (".py", ".pyc"):
            module_name = f_path.replace(path_split_char, ".")
            if module_name not in all_packages_file:
                all_packages_file.append(module_name)

    return all_packages_file
